from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import path_expand
from  pathlib import Path
import os

class Git:

    @staticmethod
    def root (path):
        path = path.replace("file:", "")
        path = os.path.abspath(path)
        directory = os.path.dirname(path)
        r = Shell.run(f"cd {directory} && git rev-parse --show-toplevel").strip()
        return r

    @staticmethod
    def name(path):
        path = path.replace("file:", "")
        path = os.path.abspath(path)
        repo = Git.repo(path)
        repo = repo.replace("git@github.com:", "").replace(".git", "")
        repo = repo.replace("https://github.com:", "")
        return repo

    @staticmethod
    def repo(path):
        path = path.replace("file:", "")
        path = os.path.abspath(path)
        directory = os.path.dirname(path)
        r =  Shell.run(f"cd {directory} && git config --get remote.origin.url").strip()
        return r

    @staticmethod
    def branch(path):
        path = path.replace("file:", "")
        path = os.path.abspath(path)
        directory = os.path.dirname(path)
        return Shell.run(f"cd {directory} && git rev-parse --abbrev-ref HEAD").strip()

    @staticmethod
    def filename(path):
        name = Git.name(path)
        repo = Git.repo(path)
        root = Git.root(path)
        path = path.replace("file:", "")
        path = os.path.abspath(path)
        path = path.replace(root, "")
        return path

    @staticmethod
    def blob(path):
        name = Git.name(path)
        filename = Git.filename(path)
        branch = Git.branch(path)
        return f"https://github.com/{name}/blob/{branch}/{filename}"

    @staticmethod
    def contributions_by_line():
        r = Shell.run('git ls-files | while read f; do git blame -w -M -C -C --line-porcelain "$f" '
                      '| grep \'^author \'; done | sort -f | uniq -ic | sort -nr')
        r = r.replace(" author ", " ")
        result = {}
        i = 0
        for line in r.splitlines():
            count, author = line.strip().split(" ", 1)
            i = i + 1
            result[i] = {
                'author': author,
                'count': count
            }
        return result

    @staticmethod
    def comitters():
        r =  Shell.run("git log --all --format='%an <%ae>' -- `git grep -l \"search string\"` | sort -u").strip()
        return r



from github import Github
from cloudmesh.configuration.Config import Config
from pprint import pprint
import requests
from textwrap import dedent
from pathlib import Path
import time
import csv
from cloudmesh.common.util import readfile


class Manager(object):

    def __init__(self, organization="cloudmesh-community"):
        config = Config()

        g = Github(config["cloudmesh.github.user"],
                   config["cloudmesh.github.password"])

        if organization != "cloudmesh-community":
            raise ValueError(
                "currently we support only organization cloudmesh-community")

        self.org = g.get_organization(organization)
        self.ta_team = self.org.get_team(2631498)

    def list(self, match=None):
        for r in self.org.get_repos():
            if match is None:
                print(r.name, r.description)
            else:
                name = r.name or ""
                description = r.description or ""
                if match in name or match in description:
                    print(r.name, r.description)

    def create_repo(self,
                    firstname=None,
                    lastname=None,
                    name=None,
                    community=None,
                    semester="fa19",
                    githubid=None
                    ):

        description = f"{firstname} {lastname}"
        repo = self.org.create_repo(name,
                                    description=description,
                                    license_template="apache-2.0")
        readme = dedent(f"""
                    ---
                    owner:
                      firstname: "{firstname}"
                      lastname: "{lastname}"
                      hid: "{name}"
                      community: "{community}"
                      semester: "{semester}"
                    """).strip()

        print(readme)
        print("Add README.yml")
        repo.create_file("README.yml",
                         "Create the Readme.yml",
                         readme,
                         branch="main")

        print("Add .gitignore")

        # bug find file within distribution

        with open(Path(".gitignore").resolve()) as file:
            gitignore = file.read()

        repo.create_file(".gitignore", "create the .gitignore", gitignore,
                         branch="main")

        try:
            repo.add_to_collaborators(githubid, permission="write")
        except Exception as e:
            pass
        self.ta_team.add_to_repos(repo)
        self.ta_team.set_repo_permission(repo, "write")

    def create_repos(self, filename=None):

        with open(filename, newline='') as csvfile:
            reader = csv.DictReader(csvfile, delimiter=',')

            for row in reader:
                firstname = row['firstname']
                lastname = row['lastname']
                githubid = row['githubid']
                community = row['community']
                semester = row['semester']
                name = row['repo']
                print(f"Create: {name} {firstname} {lastname} {githubid}")
                self.create_repo(
                    firstname=firstname,
                    lastname=lastname,
                    name=name,
                    community=community,
                    semester=semester,
                    githubid=githubid
                )

    def issue(self, repos=None, title=None, file=None):
        pprint(repos)
        for repo in repos:
            if file is not None:
                content = readfile(file).strip()

                if title is None:
                    title = content.splitlines()[0]
                    title = title.replace("#", "").strip()

                repository_obj = self.org.get_repo(repo)
                repository_obj.create_issue(title=title, body=content)



from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from cloudmesh.common.util import writefile
from cloudmesh.common.console import Console
import json
import os

class Gh:

    def __init__(self):
        self.cache = path_expand("~/.cloudmesh/issuelist.json")
        self.issue_list = None

    def cache_delete(self):
        if self.cache_exists():
            os.remove(self.cache)

    def cache_exists(self):
        return os.path.isfile(self.cache)

    def cache_load(self):
        content = readfile(self.cache)
        # do something with the content and load it into _issue_list
        self.issue_list = eval(content)

    def cache_save(self):
        # safe json as string  with json.dumps
        content = json.dumps(self.issue_list, indent=2)
        writefile(filename=self.cache, content=content)

    def repos_in_dir(self, directory="."):
        _directory = path_expand(directory)
        repos = [name for name in os.listdir(_directory) if os.path.isdir(name) and os.path.isdir(f"{name}/.git")]
        return repos

    def run(self, command, path="."):
        directory = path_expand(path)
        command = f'cd {directory} && {command}'
        try:
            r = Shell.run(command)
        except Exception as e:
            print(e)

    def issues_find(self, assignee, path=["."]):
        _issues = []
        for p in path:
            for a in assignee:
                r = self.issues_from_repos(assignee=a, path=[p])
                _issues.append(r)
        return _issues

    def issues_from_repos(self, assignee=None, path=["."]):
        _issues = []
        for p in path:
            r = self.issues(assignee=assignee, path=p)
            _issues.append(r)
        self.issue_list = _issues
        return _issues

    def issues(self, assignee=None, path="."):
        r = None
        directory = path_expand(path)
        if assignee is None:
            parameter = ""
        else:
            parameter = f'--assignee "{assignee}"'
        command = f'cd {directory} && gh issue list {parameter} --json=title,assignees,url,labels'
        print (command)

        try:
            r = Shell.run(command)
            r = json.loads(r)
        except Exception as e:
            print(e)
        return r

    def issues_to_table(self, entries):

        try:
            entry = entries[0]
            repo = os.path.dirname(entry['url'])
            repo_url = os.path.dirname(repo)
            repo_name = repo_url.replace("https://github.com/", "")
            name = repo_name
        except:
            name = "???"

        result = [
            f'<br><b>{name}</b><br><br>',
            '<table frame="border" rules="all" border-spacing="30px";>']
        line = f'<tr><th> repo </th><th> url </th><th> title </th><th> assignees</th><th> labels</th></tr>'
        result.append(line)


        if entries:
            for entry in entries:
                url = entry['url']
                title = entry['title']
                assignees = entry['assignees']
                labels = entry['labels']
                n = os.path.basename(url)
                repo = os.path.dirname(url)
                repo_url = os.path.dirname(repo)
                repo_name = repo_url.replace("https://github.com/", "")
                entry['repo'] = f'<a href="{repo_url}"> {repo_name} </a>'
                entry['url'] = f'<a href="{url}"> {n} </a>'
                entry['title'] = f'<a href="{url}"> {title} </a>'
                entry['assignees'] = ",".join(
                    [f'<a href="https://github.com/{assignee["login"]}"> {assignee["name"] or assignee["login"]} </a>'
                     for assignee in assignees])
                if len(labels) > 0:
                    entry['labels'] = ",".join([label['name'].replace("PRIORITY ", "") for label in labels])
                else:
                    entry['labels'] = ""
                line = f'<tr>'\
                       f'<td style="padding:5px"> {entry["repo"]} </td>'\
                       f'<td style="padding:5px"> {entry["url"]} </td>'\
                       f'<td style="padding:5px"> {entry["assignees"]}</td>' \
                       f'<td style="padding:5px"> {entry["title"]} </td>' \
                       f'<td style="padding:5px">{entry["labels"]}<td>'\
                       '</tr>'
                result.append(line)
            result.append("</table>")
            result = "\n".join(result)
        else:
            result = None
        return result


import glob
import json
import os
import re
import subprocess
from pprint import pprint

from cloudmesh.common.Shell import Shell
from cloudmesh.common.console import Console
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.systeminfo import os_is_linux
from cloudmesh.common.systeminfo import os_is_mac
from cloudmesh.common.systeminfo import os_is_windows
from cloudmesh.common.util import banner
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from cloudmesh.common.util import str_bool
from cloudmesh.common.util import writefile
from cloudmesh.git.api.manager import Manager
from cloudmesh.git.copy import copy_dir
from cloudmesh.git.gh import Gh
from cloudmesh.git.Git import Git
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command, map_parameters
from cloudmesh.common.Printer import Printer


class GitCommand(PluginCommand):

    # noinspection PyUnusedLocal
    @command
    def do_git(self, args, arguments):
        """
        ::

          Usage:
                git contribution
                git committers
                git create issue --repo=REPO --file=FILE [--title=TITLE] [--org=ORG]
                git create repository FIRSTNAME LASTNAME GITHUBID [--org=ORG]
                git create repository --file=FILE [--org=ORG]
                git list all [--exclude=ORG]
                git list --org=ORG [MATCH]
                git copy FROM TO DIRS... [--move=TMP]
                git set ssh [DIRS]
                git --refresh
                git clone all [--force=no]
                git pull all
                git issues [--repo=REPO] [--assignee=ASSIGNEE] [--format=HTML] [--out=a.html] [--refresh]

          This command does some useful things.

          Arguments:
              FILE   a file name
              ORG    [default: cloudmesh-community]
              MATCH  is a string that must occur in the repo name or description
              --file=FILE   specify the file
              --repo=REPO   a parameterized list of repo links. If cloudmesh- is the prefix it will be
                            automatically replaced by the repo url. REPO is a parameterized list with
                            comma separated so that multiple repos can be used.
                            If REPO is a filename, each line specifies a repo. the cloudmesh prefix
                            replacement is applied on each line
                            If repo is . or the parameter is omitted, all directories in the current
                            directory are used to list their issues
             --assignee=ASSIGNEE  a list of assignees so only issues for these users are displayed
                                  if it is ommitted all issues for assignees are displayed.
             --refresh  only download the GitHub issue list if --refresh is uses, otherwise it uses a local cache in
                        ~/.cloudmesh/issuelist.json

          Options:
              --force=no    pull the repository if it already exists in current working directory [default: no]

          Description:

                The organization is set by default to
                cloudmesh-community

                git --refresh
                    finds all organizations and repositories the current user belongs to
                    redirects automatically to ~/cloudmesh/git/repo-list.txt

                git clone all [--force=no]
                    uses all organizations and repositories of the user and
                    clones them into the current directory, while making each
                    organization in its own subdirectory
                    uses automatically ~/cloudmesh/git/repo-list.txt
                    which can be created with cms git list all.
                    if force is yes then it pulls preexisting directories

                git set ssh
                    switches the repository to use ssh

                git list --org=ORG
                    lists the repos of the organization

                git list all [--exclude=ORG]
                    gets info of all repos of the current in user. It puts
                    the result in ~/.cloudmesh/gitcache.txt.
                    To exclude an organization, add it to the end of exclude
                    parameter

                git create issue --repo=REPO FILE
                   Create an issue in the given repos.
                   Note that the repos is a string defined as
                   cloudmesh.Parameter, E.g. fa19-516-[100-103,105]
                   effects the repos ending with 100, 101, 102,
                   103, and 105

                   The bundle is defined in cloudmesh-installer

                git create repo NAME FIRSTNAME LASTNAME GITHUBID
                    creates the repo

                git create repo --file=repos.csv
                    creates repos from a file in csv format
                    the format in th csv file is

                    reponame,lastname,firstname,githubid

                git copy FROM TO
                    copies a directory from one repo to the other

                git pull all
                    git pulls all directories and subdirectories of
                    current working directory

                git issuelist
                    creates html file of all issues assigned to logged-in
                    user. assumes that the user is standing in cm
                    directory

          Examples:

               git copy FROM TO

                    git copy cloudmesh/cloudmesh-cloud cloudmesh/cloudmesh-db admin

                    creates a script move.sh that copies the directory admin
                    with history to the cloudmesh-db repo into a folder move

                    from there you can use git mv to place the content where you
                    like. The reason we put it in move is that there may be another
                    dir already in it with tha name.

               git list Park
                  Lists all repos with the name Park in it or its description

               git list fa19-523
                    Lists all repos with the string  fa19-523 in its name


        """
        # arguments.FILE = arguments['--file'] or None

        map_parameters(arguments,
                       'fetch',
                       'move',
                       'repo',
                       'file',
                       'title',
                       'assignee'
                       )
        move = arguments.move or "move"

        # VERBOSE(arguments)

        # if arguments.FILE:
        #    print("option a")
        #    m.list(path_expand(arguments.FILE))
        #

        if arguments.committers:
            r = Git.comitters()
            #print(Printer.write(r))
            print (r)
            return ""

        elif arguments.contribution:
            r = Git.contributions_by_line()
            print(Printer.write(r))
            return ""

        elif arguments.list and arguments.all:

            command = "gh api  /user/memberships/orgs"
            r = Shell.run(command)
            # print(r)

            result = json.loads(r)

            result2 = json.dumps(result, indent=2)
            # pprint(result2)
            exclude = Parameter.expand(arguments["--exclude"]) or []
            organizations = []

            for entry in result:
                url = entry["organization_url"]
                name = os.path.basename(url)
                if name not in exclude:
                    organizations.append(name)

            # pprint(organizations)
            repos = []
            for org in organizations:
                command = f"gh repo list {org} -L 1000"
                r = Shell.run(command)
                count = len(r.splitlines())
                Console.msg(f"List repos for {org}. Found {count}")
                lines = [x.split()[0] for x in r.splitlines()]
                repos = repos + lines

            pprint(repos)

            filename = path_expand("~/.cloudmesh/git_cache.txt")
            writefile(filename, "\n".join(repos))
            Console.ok(f'\nWritten list of repos to {filename}')

        # elif arguments["list"]:

        #    '''m = Manager()

        #    m.list(arguments.MATCH)'''

        elif arguments.list and arguments["--org"]:
            command = f'gh api  /orgs/{arguments["--org"]}/repos'
            r = Shell.run(command)
            # print(r)

            result = json.loads(r)

            result2 = json.dumps(result, indent=2)
            # pprint(result2)

            repos = []

            result2 = json.dumps(result, indent=2)
            pprint(result2)

            for entry in result:
                name = entry["full_name"]
                repos.append(name)

            # pprint(organizations)

            pprint(repos)
            filename = path_expand("~/.cloudmesh/git_cache.txt")
            writefile(filename, "\n".join(repos))
            Console.ok(f'\nWritten list of repos to {filename}')

        elif arguments.issues:

            github = Gh()

            # currently only allowing one user
            #if arguments.assignee is not None:
            #    assignee = Parameter.expand(arguments.assignee)[0]
            # TODO: assignee = Parameter.expand(arguments.assignee)
            assignee = arguments.assignee
            if arguments.repo in ['.', "cwd", None]:
                repos = github.repos_in_dir()
            elif arguments.repo in ["pi"]:
                repos = ["cloudmesh-pi-burn", "cloudmesh-pi-cluster", "cloudmesh-git"]
            elif arguments.repo in ["reu"]:
                repos = ["reu2022",
                         "cloudmesh-slurm",
                         "cloudmesh-mpi",
                         "cloudmesh-pi-burn",
                         "cloudmesh-pi-cluster",
                         "cloudmesh-git",
                         "cloudmesh-catalog",
                         "cloudmesh-common",
                         "cloudmesh-data",
                         "cloudmesh-sbatch",
                         "book",
                         "bookmanager",
                         "yamldb"]

            refresh = arguments["--refresh"] or not github.cache_exists()
            if refresh:
                github.cache_delete()
                github.issues_from_repos(path=repos, assignee=assignee)
                # TODO: github.issues_find(path=repos, assignee=assignee)
                github.cache_save()
            else:
                github.cache_load()

            tables = ""
            total = 0
            for d in github.issue_list:
                print(len(d))
                total = total + len(d)
                table = github.issues_to_table(d)
                if not table:
                    continue
                tables = tables + table + "\n"
            tables = f"Total Issues: {total}<br>" + tables
            html = path_expand("~/.cloudmesh/issuelist.html")
            print("Total Issues", total)
            writefile(html, tables)
            Shell.browser(html)

        elif arguments.clone and arguments["all"]:
            filename = path_expand("~/.cloudmesh/git_cache.txt")
            repos = readfile(filename).splitlines()
            failed_repos = []
            forcing_pull = str_bool(arguments["--force"])
            for repo in repos:
                url = f"git@github.com:{repo}.git"
                org = os.path.dirname(repo)
                name = os.path.basename(repo)
                command = f"mkdir -p {org}; cd {org}; git clone {url}"
                banner(command)
                try:
                    r = Shell.run(command)
                    Console.ok(f"Successfully cloned {repo}.")
                except subprocess.CalledProcessError as e:
                    if forcing_pull:
                        if "already exists and is not an empty directory" in str(e.output):
                            pull_command = f"cd {org}; cd {name}; git pull"
                            banner(pull_command)
                            try:
                                r2 = Shell.run(pull_command)
                                Console.ok(f"Pulled {repo} since it already exists.")
                            except subprocess.CalledProcessError as e2:
                                Console.error(f"Failed to pull {repo}. Continuing...")
                                failed_repos.append(repo)
                                continue
                    else:
                        if "already exists and is not an empty directory" in str(e.output):
                            Console.ok(f"Skipping {repo} because it already exists.")
                        else:
                            Console.error(f"Failed to clone {repo}. Continuing...")
                            failed_repos.append(repo)
                    continue
            if failed_repos:
                Console.error(f"These repos failed to clone:\n")
                for failed_repo in failed_repos:
                    print(f'{failed_repo}\n')

        elif arguments.create and arguments.repo is not None:
            """
            git create issue --repo=REPO --title=TITLE --file=FILE [--org=ORG]
            """
            m = Manager()

            file = arguments.file
            title = arguments.title
            repo = arguments.repo
            repos = Parameter.expand(repo)
            m.issue(repos=repos, title=title, file=file)

        elif arguments.repository and arguments.file and not arguments.issue:

            m = Manager()
            filename = arguments.file
            m.create_repos(filename=filename)

        elif arguments.ssh and arguments.set:

            dirs = arguments["DIRS"] or "."
            org = "get the org from the current dir in .git"
            repo = "get the repo from the current dir in .git"

            for d in dirs:
                if d == ".":
                    location = ""
                else:
                    location = "cd {d}; "
            os.system(f"{location} git remote set-url origin git@github.com:{org}/{repo}.git")

        elif arguments.pull and arguments.all:
            for path in glob.glob(f'./**/', recursive=True):
                command = f"git -C {path} pull"
                banner(command)
                os.system(command)

        elif arguments.copy:
            dirs = arguments.DIRS
            original = arguments.FROM
            destination = arguments.TO
            move = arguments.move

            copy_dir(original=original,
                     destination=destination,
                     directories=dirs,
                     move=move)

        return ""



import glob
import os
import textwrap

from cloudmesh.common.console import Console


# inspired from
# https://gist.githubusercontent.com/trongthanh/2779392/raw/6aa7ef47ed3c1a586fafd1c8b74adcf4216e52ce/gistfile1.sh


#
# cms git copy cloudmesh/cloudmesh-cloud cloudmesh/cloudmesh-admin cloudmesh/admin cloudmesh/check cloudmesh/management cloudmesh/source cloudmesh/start cloudmesh/stop
#

def copy_dir(original="cloudmesh/cloudmesh-cloud",
             destination=None,
             directories=None,
             move=None,
             branch=None):
    def _run(script):
        for command in script.splitlines():
            if command not in [""]:
                print(command)
                os.system(command)
            else:
                print()

    if None in [original, destination, directories]:
        Console.error("A parameter is missing")
        return ""
    move = move or "move"
    dir_original = os.path.basename(original)
    dir_destination = os.path.basename(destination)

    branch = branch or f"from-{dir_original}"
    cache = "cache"

    dirs = ' '.join(directories)

    script = "rm -rf tmp"
    if not os.path.isdir(f"{cache}"):
        script += textwrap.dedent(f"""
        rm -rf {cache}
        mkdir {cache}
        cd {cache}; git clone git@github.com:{original}.git
        cd {cache}; git clone git@github.com:{destination}.git
        """)
    script += textwrap.dedent(f"""
        cp -r {cache} tmp
        cd tmp/{dir_original}; mkdir {move}
    """)

    _run(script)

    script = ""
    #
    # collect directories to {move}
    #
    os.chdir(f"tmp/{dir_original}")

    move_dirs = set()
    for directory in directories:
        dirname = os.path.dirname(directory)
        move_dirs.add(f"{dirname}")

    for directory in move_dirs:
        script += "git remote rm origin\n"
        script += f"mkdir -p {move}/{directory}\n"
        script += f"git add  {move}/{directory}\n"

    for directory in directories:
        script += f"git mv {directory} move/{directory}\n"

    entries = glob.glob("**")
    print(entries)
    entries.remove(f"{move}")

    for entry in entries:
        script += f"rm -rf {entry}\n"

    script += textwrap.dedent(f"""
    git filter-branch --subdirectory-filter --prune-empty {move} -- -- all 
    git commit -m "Directories moved: {dirs}" .
    git gc --aggressive
    git prune
    git clean -df
    """)

    _run(script)

    #
    # Prepare the destination
    #
    os.chdir(f"../{dir_destination}")

    script = textwrap.dedent(f"""
    git remote add {branch} ../{dir_original}
    git fetch {branch}
    git branch {branch} remotes/{branch}/main
    git merge {branch} --allow-unrelated-histories

    git remote rm {branch}
    git branch -d {branch}

    pwd

    ls 

    """)

    _run(script)

    print("Now do:")
    print()
    print(f"cd tmp/{dir_destination}; git push origin main")
    print()
